#include "formula.h"

template<class ModalValueType>
Formula<ModalValueType>::Formula(){
	number_of_variables = 0;
	variables.set_empty_key(-1);            //
	variables_back.set_empty_key(-1);       //
	data d;
	modal_variables.set_empty_key(d); 
	modal_variables_back.set_empty_key(-1); //
	Premise<ModalValueType> p;
	rule_cache.set_empty_key(p);
}

template<class ModalValueType>
bdd Formula<ModalValueType>::variable(int n){
	if(variables.count(n) == 0){
		variables[n] = number_of_variables;
		variables_back[number_of_variables] = n;
		number_of_variables++;
	}
	return bdd_ithvarpp(variables[n]);
}

template<class ModalValueType>
bdd Formula<ModalValueType>::modal_var(bdd *b, ModalValueType &n){
	data d(n, b);
	if(modal_variables.count(d) == 0){
		modal_variables[d] = number_of_variables;
		modal_variables_back[number_of_variables] = d;
		number_of_variables++;
	}
	return bdd_ithvarpp(modal_variables[d]);
}


template<class ModalValueType>
void Formula<ModalValueType>::print_back_vars(){
	cout << setw(10) << "BDD number   " << setw(10) << "Input Number" << endl;
	for(int i=0; i < 1000; i++)
		if(variables_back.count(i) > 0 )
			cout <<  setw(10) << i << setw(10) << variables_back[i] << endl;
	cout << endl;
}

void sat_handler(char *varset, int size){
	SatisfyingAssignment s(size, varset);
	sat_ass_stack.push_back(s);
}

template<class ModalValueType>
bool Formula<ModalValueType>::check_satisfiability(bdd b){
	bool sat = false; // handle the case of no satisfying assignments.
	//bdd_printdot(b);
	bdd_allsat(b, sat_handler); // pushes satisfying assignment onto stack
	vector<SatisfyingAssignment> sat_ass = sat_ass_stack; 
	sat_ass_stack.clear();	
	for(unsigned int i=0; i < sat_ass.size() && !sat; i++){ // we break if proved sat
		sat = apply_rules(sat_ass[i]);
	}
	return sat;
}

template<class ModalValueType>
bool Formula<ModalValueType>::apply_rules(SatisfyingAssignment& sat){
	bool result = true;
	
	int no_of_positive = 0;
	int no_of_negative = 0;
	bdd **underlying_formulae;
	ModalValueType *positive_modal_indices, *negative_modal_indices;

	// We first find the size of the required arrays.
	for(int v=0; v < sat.get_size(); v++){
		if(sat.get_array_i(v)==1 && modal_variables_back.count(v)) // if true and represents a modal formula.
			no_of_positive++;
		else if(sat.get_array_i(v)==0 && modal_variables_back.count(v)) // if false and ...
			no_of_negative++;
	}
	// cout << no_of_positive + no_of_negative << endl;
	if(no_of_positive + no_of_negative > 0){
	
		// We then initialize our new arrays.
		underlying_formulae = new bdd* [no_of_positive+no_of_negative];
	
		positive_modal_indices = new ModalValueType[no_of_positive];
		negative_modal_indices = new ModalValueType[no_of_negative];

		// We then fill our underlying formulae and modal value arrays.
		int i=0, j=0;
		for(int v=0; v < sat.get_size(); v++){
			if(sat.get_array_i(v)==1 && modal_variables_back.count(v)){
				underlying_formulae[i] = (modal_variables_back[v]).argument;
				positive_modal_indices[i] = (modal_variables_back[v]).value;
				i++;
			}
			else if(sat.get_array_i(v)==0 && modal_variables_back.count(v)){ //if false and represents a modal formula.
				underlying_formulae[no_of_positive+j] = (modal_variables_back[v]).argument;
				negative_modal_indices[j] = (modal_variables_back[v]).value;
				j++;
			}
		}

		// Construct a GML premise
		Premise<ModalValueType> premise(no_of_positive, no_of_negative, positive_modal_indices, negative_modal_indices);
		// Check the rule cache
		if(rule_cache.count(premise)){ // if in cache
			SetOfConclusions concs = rule_cache[premise];
			for(int i = 0; i < concs.number_of_conclusions() && result; i++){
				bdd b = concs.get_jth_conclusion(underlying_formulae, i);
				result = check_satisfiability(b);
			}
		} else {
			result = enumerate_rules(premise, underlying_formulae);
		}

		delete [] underlying_formulae;
		delete [] positive_modal_indices;
		delete [] negative_modal_indices; 
	}
	
	return result;
}

template<class ModalValueType>
bool Formula<ModalValueType>::enumerate_rules(Premise<ModalValueType>& prem, bdd** underlying_formulae){
	bool result = true;
	vector<vector<bool> > valid_nodes;
	vector<pair<vector<bool>, int> > anti_nodes;
	vector<vector<bool> > node_queue;
	int total_valuations = prem.get_total_valuations();

	glp_prob *problem = glp_create_prob();

	// Parameters while solving with glpk
	glp_iocp* parameters = new glp_iocp;
	glp_init_iocp(parameters);
	parameters->presolve = GLP_ON;

	parameters->msg_lev = GLP_MSG_ERR;
	//parameters->msg_lev = GLP_MSG_ALL;
	
	load_linear_program(problem, prem);
	
	vector<bool> zero(total_valuations, false);
	
	if (run_solver(problem, parameters, false, zero, -1)) {
		if (run_solver(problem, parameters, true, zero, 0))
			result = false;
		else {
			node_queue.push_back(zero);
			//int i = 0;
			while(result && node_queue.size()) {
				//cout << "outer space: " << i++ << "  runs: " << test_counter << "  queue: " << node_queue.size() << "  sols: " << valid_nodes.size() << "  antisols: " << anti_nodes.size() << endl;
				result = enum_rules_intern(problem, parameters, prem, underlying_formulae, node_queue, valid_nodes, anti_nodes);
			}
			if (result && !rule_cache.count(prem)) {
				rule_cache[prem] = SetOfConclusions(prem.get_n() + prem.get_m(), valid_nodes);
			}
			//cout << "runs: " << test_counter << "  sols: " << valid_nodes.size() << "  antisols: " << anti_nodes.size() << endl;
		}
	}

	//Clean-up
	delete parameters;
	glp_delete_prob(problem);
	
	return result;
}

template<class ModalValueType>
bool Formula<ModalValueType>::enum_rules_intern(glp_prob* problem, glp_iocp* parameters, Premise<ModalValueType>& prem, bdd** underlying_formulae,
												vector<vector<bool> >& node_queue, vector<vector<bool> >& valid_nodes, vector<pair<vector<bool>, int> >& anti_nodes){
	bool result = true;
	vector<vector<bool> > new_node_queue;
	int total_valuations = prem.get_total_valuations();

	for (unsigned int k = 0; k < node_queue.size() && result; k++) {
		vector<bool> current_node = node_queue[k];

		int power = total_valuations-1;
		while(power >= 0 && !current_node[power])
			power--;
		for(int i = power+1; i < total_valuations && result; i++){
			vector<bool> new_node(current_node);
			new_node[i] = true;

			// Check node isn't superset of something already valid
			bool isSubnode = false;
			for(unsigned int i1=0; i1 < valid_nodes.size() && !isSubnode; i1++){
				bool isSub = true;
				for(int j1=0; j1 < total_valuations; j1++){
					if (valid_nodes[i1][j1] && !new_node[j1]) {
						isSub = false;
						break;
					}
				}
				if (isSub) isSubnode = true;
			}
			// Check node isn't superset of something of an anti-solution
			for(unsigned int i1=0; i1 < anti_nodes.size() && !isSubnode; i1++){
				bool isSub = true;
				int idx = anti_nodes[i1].second;
				vector<bool> antinode = anti_nodes[i1].first;
				for(int j1=0; j1 < total_valuations; j1++){
					if ((j1 <= idx && antinode[j1] != new_node[j1]) || (j1 > idx && antinode[j1] && !new_node[j1])) {
						isSub = false;
						break;
					}
				}
				if (isSub) isSubnode = true;
			}

			if (!isSubnode) {
				if (run_solver(problem, parameters, false, new_node, i)) {
					if (run_solver(problem, parameters, true, new_node, 0)) {
						valid_nodes.push_back(new_node);
						//cout << "solution found " << valid_nodes.size() << endl;
						//for (int l = 0; l < total_valuations; l++)
						//	cout << "   " << new_node[l] << endl;
						vector<vector<bool> > singleton;
						singleton.push_back(new_node);
						SetOfConclusions temp = SetOfConclusions(prem.get_n() + prem.get_m(), singleton);
						bdd b = temp.get_jth_conclusion(underlying_formulae, 0);
						result = check_satisfiability(b);
					} else {
						new_node_queue.push_back(new_node);
					}
				}
				else {
					pair<vector<bool>, int> antipair = pair<vector<bool>, int>(new_node, i);
					anti_nodes.push_back(antipair);
				}
			}
		}
	}
	node_queue = new_node_queue;
	return result;
}	

template<class ModalValueType>
bool Formula<ModalValueType>::run_solver(glp_prob *problem, glp_iocp* parameters, bool strict, vector<bool> current_node, int index){
	test_counter++;

	// Modify glp problem to match node
	for(unsigned int i = 0; i < current_node.size(); i++){
		if(current_node[i])
			glp_set_row_bnds(problem, i+1, GLP_LO, 1.0, 0.0); // > 0
			//+1 since valuations go zero to 2^(n+m)-1, glpk rows go 1 to 2^(n+m).
		else if (strict || static_cast<int>(i) <= index)
			glp_set_row_bnds(problem, i+1, GLP_UP, 0.0, 0.0); // leq 0
		else
			glp_set_row_bnds(problem, i+1, GLP_FR, 0.0, 0.0); // don't care
	}

	if (!strict) {
		parameters->tm_lim = 1000;
	} else {
		parameters->tm_lim = 1000000000;
	}

	// Solve the LP
	int result = glp_intopt(problem, parameters);
		
	if(result == 0 || result == GLP_ENOPFS) { //if no errors from solver or a relaxtion feasibility error
		if(glp_mip_status(problem)==GLP_OPT || glp_mip_status(problem)==GLP_FEAS){ // if feasible
			return true;
		}
		else {
			return false;
		}
	}
	else if (!strict && result == GLP_ETMLIM) { // if the solver timed out for non-strict problems
		//cout << "timeout" << endl;
		return true;
	} else { // if error but not a feasibility error
		cout << "glpk error: " << result << endl;
		exit(1);
	}
}


template class Formula<int>;
template class Formula<Rational>;
